@@ -28,51 +28,42 @@ def stream_unzip(zipfile_chunks, password=None, chunk_size=65536):
         # - _get_num: returns a single `bytes` of a given length
         # - _return_unused: puts "unused" bytes "back", to be retrieved by a yield/get call
 
-        prev_chunk = b''
         chunk = b''
         offset = 0
         it = iter(iterable)
 
         def _yield_all():
-            nonlocal prev_chunk, chunk, offset
+            nonlocal chunk, offset
 
             while True:
-                if not chunk:
+                if offset == len(chunk):
                     try:
                         chunk = next(it)
                     except StopIteration:
                         break
-                prev_offset = offset
-                prev_chunk = chunk
+                    else:
+                        offset = 0
                 to_yield = min(len(chunk) - offset, chunk_size)
-                offset = (offset + to_yield) % len(chunk)
-                chunk = chunk if offset else b''
-                yield prev_chunk[prev_offset:prev_offset + to_yield]
+                offset = offset + to_yield
+                yield chunk[offset - to_yield:offset]
 
         def _yield_num(num):
-            nonlocal prev_chunk, chunk, offset
-
-            while num:
-                if not chunk:
-                    chunk = next_or_truncated_error(it)
-                prev_offset = offset
-                prev_chunk = chunk
-                to_yield = min(num, len(chunk) - offset, chunk_size)
-                offset = (offset + to_yield) % len(chunk)
-                chunk = chunk if offset else b''
-                num -= to_yield
-                yield prev_chunk[prev_offset:prev_offset + to_yield]
-
-        def _get_num(num):
-            return b''.join(chunk for chunk in _yield_num(num))
-
-        def _return_unused(num_unused):
             nonlocal chunk, offset
 
-            if num_unused and not chunk:
-                chunk = prev_chunk
-                offset = len(chunk)
+            while num:
+                if offset == len(chunk):
+                    chunk = next_or_truncated_error(it)
+                    offset = 0
+                to_yield = min(num, len(chunk) - offset, chunk_size)
+                offset = offset + to_yield
+                num -= to_yield
+                yield chunk[offset - to_yield:offset]
 
+        def _get_num(num):
+            return b''.join(_yield_num(num))
+
+        def _return_unused(num_unused):
+            nonlocal offset
             offset -= num_unused
 
         return _yield_all, _get_num, _return_unused
@@ -132,7 +123,7 @@ def stream_unzip(zipfile_chunks, password=None, chunk_size=65536):
 
         def parse_extra(extra):
             extra_offset = 0
-            while extra_offset < len(extra):
+            while extra_offset <= len(extra) - 4:
                 extra_signature = extra[extra_offset:extra_offset+2]
                 extra_offset += 2
                 extra_data_size, = Struct('<H').unpack(extra[extra_offset:extra_offset+2])
@@ -159,16 +150,15 @@ def stream_unzip(zipfile_chunks, password=None, chunk_size=65536):
             key_0 = 305419896
             key_1 = 591751049
             key_2 = 878082192
-
-            def crc32(ch, crc):
-                return ~zlib.crc32(bytes([ch]), ~crc) & 0xFFFFFFFF
+            crc32 = zlib.crc32
+            bytes_c = bytes
 
             def update_keys(byte):
                 nonlocal key_0, key_1, key_2
-                key_0 = crc32(byte, key_0)
+                key_0 = ~crc32(bytes_c((byte,)), ~key_0) & 0xFFFFFFFF
                 key_1 = (key_1 + (key_0 & 0xFF)) & 0xFFFFFFFF
                 key_1 = ((key_1 * 134775813) + 1) & 0xFFFFFFFF
-                key_2 = crc32(key_1 >> 24, key_2)
+                key_2 = ~crc32(bytes_c((key_1 >> 24,)), ~key_2) & 0xFFFFFFFF
 
             def decrypt(chunk):
                 chunk = bytearray(chunk)
@@ -177,7 +167,7 @@ def stream_unzip(zipfile_chunks, password=None, chunk_size=65536):
                     byte ^= ((temp * (temp ^ 1)) >> 8) & 0xFF
                     update_keys(byte)
                     chunk[i] = byte
-                return chunk
+                return bytes(chunk)
 
             for byte in password:
                 update_keys(byte)
@@ -190,20 +180,17 @@ def stream_unzip(zipfile_chunks, password=None, chunk_size=65536):
 
             return_unused(num_unused())
 
-        def aes_decrypt_decompress(chunks, decompress, is_done, num_unused, aes_extra):
-            if aes_extra[4] not in (1, 2, 3):
-                raise InvalidAESKeyLengthError(aes_extra[4])
-
-            key_length = {1: 16, 2: 24, 3: 32}[aes_extra[4]]
-            salt_length = {1: 8, 2: 12, 3: 16}[aes_extra[4]]
-            compression = aes_extra[5:7]
+        def aes_decrypt_decompress(chunks, decompress, is_done, num_unused, key_length_raw):
+            try:
+                key_length, salt_length = {1: (16, 8), 2: (24, 12), 3: (32, 16)}[key_length_raw]
+            except KeyError:
+                raise InvalidAESKeyLengthError(key_length_raw)
 
             salt = get_num(salt_length)
             password_verification_length = 2
-            password_verification = get_num(password_verification_length)
 
             keys = PBKDF2(password, salt, 2 * key_length + password_verification_length, 1000)
-            if keys[-2:] != password_verification:
+            if keys[-password_verification_length:] != get_num(password_verification_length):
                 raise IncorrectAESPasswordError()
 
             decrypter = AES.new(
@@ -257,7 +244,7 @@ def stream_unzip(zipfile_chunks, password=None, chunk_size=65536):
             if crc_32_actual != get_crc_32_expected():
                 raise CRC32IntegrityError()
 
-        version, flags, compression_raw, mod_time, mod_date, crc_32_expected, compressed_size, uncompressed_size, file_name_len, extra_field_len = \
+        version, flags, compression_raw, mod_time, mod_date, crc_32_expected, compressed_size_raw, uncompressed_size_raw, file_name_len, extra_field_len = \
             local_file_header_struct.unpack(get_num(local_file_header_struct.size))
 
         flag_bits = tuple(get_flag_bits(flags))
@@ -290,15 +277,13 @@ def stream_unzip(zipfile_chunks, password=None, chunk_size=65536):
         if compression not in (0, 8):
             raise UnsupportedCompressionTypeError(compression)
 
-        is_zip64 = compressed_size == zip64_compressed_size and uncompressed_size == zip64_compressed_size
+        is_zip64 = compressed_size_raw == zip64_compressed_size and uncompressed_size_raw == zip64_compressed_size
         zip64_extra = get_extra_value(extra, is_zip64, zip64_size_signature, MissingZip64ExtraError, 16, TruncatedZip64ExtraError)
 
         has_data_descriptor = flag_bits[3]
-        uncompressed_size_raw, compressed_size = \
-            Struct('<QQ').unpack(zip64_extra) if is_zip64 else \
-            (uncompressed_size, compressed_size)
         uncompressed_size = \
             None if has_data_descriptor and compression == 8 else \
+            Struct('<Q').unpack(zip64_extra[:8])[0] if is_zip64 else \
             uncompressed_size_raw
 
         decompressor = \
@@ -307,7 +292,7 @@ def stream_unzip(zipfile_chunks, password=None, chunk_size=65536):
 
         decompressed_bytes = \
             weak_decrypt_decompress(yield_all(), *decompressor) if is_weak_encrypted else \
-            aes_decrypt_decompress(yield_all(), *decompressor, aes_extra) if is_aes_encrypted else \
+            aes_decrypt_decompress(yield_all(), *decompressor, key_length_raw=aes_extra[4]) if is_aes_encrypted else \
             no_decrypt_decompress(yield_all(), *decompressor)
 
         get_crc_32_expected = \
@@ -322,24 +307,39 @@ def stream_unzip(zipfile_chunks, password=None, chunk_size=65536):
 
     yield_all, get_num, return_unused = get_byte_readers(zipfile_chunks)
 
-    while True:
-        signature = get_num(len(local_file_header_signature))
-        if signature == local_file_header_signature:
-            yield yield_file(yield_all, get_num, return_unused)
-        elif signature == central_directory_signature:
-            for _ in yield_all():
-                pass
-            break
-        else:
-            raise UnexpectedSignatureError(signature)
+    def all():
+        while True:
+            signature = get_num(len(local_file_header_signature))
+            if signature == local_file_header_signature:
+                yield yield_file(yield_all, get_num, return_unused)
+            elif signature == central_directory_signature:
+                for _ in yield_all():
+                    pass
+                break
+            else:
+                raise UnexpectedSignatureError(signature)
 
-class UnzipError(ValueError):
+    for file_name, file_size, unzipped_chunks in all():
+        yield file_name, file_size, unzipped_chunks
+        for _ in unzipped_chunks:
+            raise UnfinishedIterationError()
+
+class UnzipError(Exception):
     pass
 
-class DataError(UnzipError):
+class InvalidOperationError(UnzipError):
     pass
 
-class UncompressError(UnzipError):
+class UnfinishedIterationError(InvalidOperationError):
+    pass
+
+class UnzipValueError(UnzipError, ValueError):
+    pass
+
+class DataError(UnzipValueError):
+    pass
+
+class UncompressError(UnzipValueError):
     pass
 
 class DeflateError(UncompressError):
@@ -393,10 +393,10 @@ class HMACIntegrityError(IntegrityError):
 class CRC32IntegrityError(IntegrityError):
     pass
 
-class PasswordError(UnzipError):
+class PasswordError(UnzipValueError):
     pass
 
-class MissingPasswordError(UnzipError):
+class MissingPasswordError(UnzipValueError):
     pass
 
 class MissingZipCryptoPasswordError(MissingPasswordError):
